from __future__ import annotations
from .file_system_item import FileSystemItem, FileSystemItemTypeT
from .file_item import FileItem
from .directory_item import DirectoryItem


class FileSystemItemList(list):
    """
    Helper class to work with a list of :class:`FileSystemItem`
    """

    def filter_by(self, file_system_item_type: type[FileSystemItemTypeT]):
        """
        Filters the current item list by a specific type of :class:`FileSystemItem`

        :param file_system_item_type: the requested type the filter should be applied for
        :return: the new filtered list
        """
        if not isinstance(file_system_item_type, type) and issubclass(file_system_item_type, FileSystemItem):
            raise TypeError(f'file_system_item_type must be a subclass of `{FileSystemItem.__name__}`')
        return self.__class__(filter(lambda item: isinstance(item, file_system_item_type), self))

    @property
    def dirs(self):
        """
        :return: filters the list by :class:`DirectoryItem` only
        """
        return self.filter_by(DirectoryItem)

    @property
    def files(self):
        """
        :return: filters the list by :class:`FileItem` only
        """
        return self.filter_by(FileItem)

    def has_duplicates(self) -> bool:
        """
        :return: returns True if the list has duplicate items
        """
        return len(self) != len(set(self))

    def compare_with(self, other: FileSystemItemList, ignore_order=False):
        """
        This method compares two :class:`FileSystemItemList`.

        :param other: the other :class:`FileSystemItemList` to compare with
        :param ignore_order: True if the order is not relevant and can be ignored during comparison
        :return: True if the lists are equal, False otherwise
        """
        this_list = self.copy()
        other_list = other.copy()
        if ignore_order:
            this_list = sorted(this_list, key=lambda f: f.name)
            other_list = sorted(other_list, key=lambda f: f.name)

        return this_list == other_list

    def subtract(self, other: FileSystemItemList, ignore_non_existing: bool = False):
        """
        This method subtracts two :class:`FileSystemItemList` from each other.
        :param other: the other :class:`FileSystemItemList` to subtract
        :param ignore_non_existing: True if the method should not throw an exception in case that the other list has
                                    elements that are not contained in this list
        :return: a :class:`FileSystemItemList` with elements that are are contained in this list, but not in the other
                 list
        """
        if self.has_duplicates():
            raise ValueError('this list has duplicates - can only apply the `subtract()` method for unique lists')
        if other.has_duplicates():
            raise ValueError('the provided list `other` has duplicates - can only apply the `subtract()` method for '
                             'unique lists')
        if not ignore_non_existing:
            for elem in other:
                if elem not in self:
                    raise ValueError(f'cannot subtract {elem} from {self}, because it does not exist')

        diff = set(self) - set(other)
        return self.__class__(diff)



import balder

from balderhub.fileops.lib.utils import DirectoryItem


class DirectoryRenamerFeature(balder.Feature):
    """
    Feature that allows to rename a directory within the filesystem
    """
    def rename_directory(self, source: DirectoryItem, rename_to: str) -> None:
        """
        This method renames the directory provided with ``source`` to the name given with ``rename_to``.

        :param source: the source directory that should be renamed
        :param rename_to: the new name the source directory should be named to
        """
        raise NotImplementedError()



from typing import TypeVar
import dataclasses


FileSystemItemTypeT = TypeVar('FileSystemItemTypeT', bound='FileSystemItem')

@dataclasses.dataclass
class FileSystemItem:
    """
    base class for elements that can be located inside the filesystem
    """
    name: str

    def __eq__(self, other):
        return self.__class__ == self.__class__ and self.name == other.name

    def __hash__(self):
        return hash(self.name)



import dataclasses
from .file_system_item import FileSystemItem


@dataclasses.dataclass
class DirectoryItem(FileSystemItem):
    """
    describes a directory
    """

    def __hash__(self):
        return hash(self.name)



from .directory_copyer_feature import DirectoryCopyerFeature
from .directory_creator_feature import DirectoryCreatorFeature
from .directory_deleter_feature import DirectoryDeleterFeature
from .directory_mover_feature import DirectoryMoverFeature
from .directory_renamer_feature import DirectoryRenamerFeature
from .file_system_navigator_feature import FileSystemNavigatorFeature


__all__ = [
    'DirectoryCopyerFeature',
    'DirectoryCreatorFeature',
    'DirectoryDeleterFeature',
    'DirectoryMoverFeature',
    'DirectoryRenamerFeature',
    'FileSystemNavigatorFeature'
]



import balder

from balderhub.fileops.scenarios.abstract_scenario_base import AbstractScenarioBase
from balderhub.fileops.lib.scenario_features import FileSystemNavigatorFeature, DirectoryCreatorFeature, \
    DirectoryMoverFeature, DirectoryDeleterFeature
from balderhub.fileops.lib.utils import DirectoryItem


class ScenarioDirMoveStandalone(AbstractScenarioBase):
    """
    Standalone scenario that provides test for MOVING directories.

    This standalone version also creates a test directory to move and cleans up everything afterward.
    """

    class Filesystem(AbstractScenarioBase.Filesystem):
        """
        The main device that has the filesystem by itself or allows to control it
        """
        navigator = FileSystemNavigatorFeature()
        dir_creator = DirectoryCreatorFeature()  # TODO another scenario with already provided dirs??
        dir_deleter = DirectoryDeleterFeature()  # TODO another scenario with already provided dirs??
        dir_mover = DirectoryMoverFeature()

    @balder.fixture('testcase')
    def dir_to_move(self):
        """fixture that creates and cleans up a test directory"""
        directory = DirectoryItem('dir_to_move')
        self.Filesystem.dir_creator.create_new_directory(directory.name)

        yield directory

        if directory in self.Filesystem.navigator.get_all_list_items():
            self.Filesystem.dir_deleter.delete_directory(directory)

    @balder.fixture('testcase')
    def destination_dir(self):
        """fixture that creates and cleans up a test directory to move directories in"""
        destination_dir = DirectoryItem('destination_dir')
        self.Filesystem.dir_creator.create_new_directory(destination_dir.name)

        yield destination_dir

        self.Filesystem.dir_deleter.delete_directory(destination_dir)

    def test_move_directory(self, dir_to_move: DirectoryItem, destination_dir: DirectoryItem):
        """
        Simple test that tries to move the test directory to the destination directory. It will automatically clean up
        everything it changed afterward.
        """
        items_before = self.Filesystem.navigator.get_all_list_items()
        assert dir_to_move in items_before

        assert destination_dir in items_before, f"can not find {destination_dir} in current elements {items_before}"
        assert dir_to_move in items_before, f"can not find {dir_to_move} in current elements {items_before}"

        # go into destination dir and check content there
        try:
            self.Filesystem.navigator.navigate_to(self.EXECUTE_IN_DIR.joinpath(destination_dir.name))
            elements_within_destination_before = self.Filesystem.navigator.get_all_list_items()
            assert dir_to_move not in elements_within_destination_before, \
                f"found another {dir_to_move} within the destionation {destination_dir}"
        finally:
            self.Filesystem.navigator.navigate_to(self.EXECUTE_IN_DIR)

        # now move the first directory
        self.Filesystem.dir_mover.move_directory(dir_to_move, destination_dir)

        items_after = self.Filesystem.navigator.get_all_list_items()

        assert items_before.files == items_after.files, \
            f"different files detected - before {items_before.files} / after {items_after.files}"

        assert len(items_before.dirs) == len(items_after.dirs) + 1, \
            (f"detect unexpected number of directories: {len(items_after.dirs)} (expected {len(items_before.dirs) + 1})"
             f" - before: {items_before.dirs} | after: {items_after.dirs}")

        assert dir_to_move not in items_after, f"{dir_to_move} should not be available in {items_after}"

        # make sure that content of `dir_to_move` is in `new_moved_dir`
        try:
            self.Filesystem.navigator.navigate_to(self.EXECUTE_IN_DIR.joinpath(destination_dir.name))
            elements_within_destination_after = self.Filesystem.navigator.get_all_list_items()
            assert dir_to_move in elements_within_destination_after, \
                f"found another {dir_to_move} within the destionation {destination_dir}"

            new_elements_in_dest_dir = elements_within_destination_after.subtract(elements_within_destination_before)

            assert len(new_elements_in_dest_dir) == 1, \
                f"detect unexpected number of elements in destination {destination_dir}: {new_elements_in_dest_dir}"
            assert new_elements_in_dest_dir[0] == dir_to_move, \
                (f"detect unexpected directory - expected that new directory is {dir_to_move}, but in reality it is "
                 f"{new_elements_in_dest_dir[0]}")

        finally:
            self.Filesystem.navigator.navigate_to(self.EXECUTE_IN_DIR)



import balder


class DirectoryCreatorFeature(balder.Feature):
    """
    Feature that allows to create a directory within the filesystem
    """
    def create_new_directory(self, name: str) -> None:
        """
        This method creates a new directory with the provided name

        :param name: the name of the new directory
        """
        raise NotImplementedError()



import balder

from balderhub.fileops.scenarios.abstract_scenario_base import AbstractScenarioBase
from balderhub.fileops.lib.scenario_features import FileSystemNavigatorFeature, DirectoryCreatorFeature, \
    DirectoryDeleterFeature
from balderhub.fileops.lib.utils import DirectoryItem


class ScenarioDirDeleteStandalone(AbstractScenarioBase):
    """
    Standalone scenario that provides test for DELETING directories.

    This standalone version also creates a test directory to delete.
    """

    class Filesystem(AbstractScenarioBase.Filesystem):
        """
        The main device that has the filesystem by itself or allows to control it
        """
        navigator = FileSystemNavigatorFeature()
        dir_creator = DirectoryCreatorFeature()   # TODO another scenario with already provided dirs??
        dir_deleter = DirectoryDeleterFeature()

    @balder.fixture('testcase')
    def dir_to_delete(self):
        """fixture that creates and cleans up a test directory"""
        directory = DirectoryItem('dir_to_delete')
        self.Filesystem.dir_creator.create_new_directory(directory.name)
        return directory

    def test_delete_empty_directory(self, dir_to_delete):
        """
        Simple test that tries to delete the test directory.
        """
        items_before = self.Filesystem.navigator.get_all_list_items()

        assert dir_to_delete in items_before, \
            f"can not find the {dir_to_delete} that should be deleted in {items_before}"

        self.Filesystem.dir_deleter.delete_directory(dir_to_delete)

        items_after = self.Filesystem.navigator.get_all_list_items()

        assert items_before.files == items_after.files, \
            f"the files should not change - before: {items_before.files} / after: {items_after.files}"

        assert len(items_after.dirs) == len(items_before.dirs) - 1, \
            f"detect unexpected number of directories: {len(items_after.dirs)} (expected {len(items_before.dirs) - 1})"

        dirs_deleted = items_before.dirs.subtract(items_after.dirs)
        assert len(dirs_deleted) == 1, (f"detect unexpected number of deleted directories: {dirs_deleted} "
                                        f"(expected `['{dir_to_delete}']`)")

        assert dirs_deleted[0] == dir_to_delete, f"detect unexpected directory: {dirs_deleted[0]}"

    #def test_delete_directory_with_other_dirs(self):
    #    pass



import balder

from balderhub.fileops.scenarios.abstract_scenario_base import AbstractScenarioBase
from balderhub.fileops.lib.scenario_features import FileSystemNavigatorFeature, DirectoryCreatorFeature, \
    DirectoryCopyerFeature, DirectoryDeleterFeature
from balderhub.fileops.lib.utils import DirectoryItem


class ScenarioDirCopyStandalone(AbstractScenarioBase):
    """
    Standalone scenario that provides test for COPYING directories.

    This standalone version also creates a test directory to copy and cleans up everything afterward.
    """
    class Filesystem(AbstractScenarioBase.Filesystem):
        """
        The main device that has the filesystem by itself or allows to control it
        """
        navigator = FileSystemNavigatorFeature()
        dir_creator = DirectoryCreatorFeature()    # TODO another scenario with already provided dirs??
        dir_deleter = DirectoryDeleterFeature()  # TODO another scenario with already provided dirs??
        dir_copyer = DirectoryCopyerFeature()

    @balder.fixture('testcase')
    def dir_to_copy(self):
        """fixture that creates and cleans up a test directory"""
        directory = DirectoryItem('dir_that_should_be_copied')
        self.Filesystem.dir_creator.create_new_directory(directory.name)
        yield directory
        self.Filesystem.dir_deleter.delete_directory(directory)

    def test_copy_directory(self, dir_to_copy):
        """
        Simple test that tries to copy the test directory. It will automatically clean up everything it changed
        afterward.
        """
        items_before = self.Filesystem.navigator.get_all_list_items()

        # now copy the first directory
        new_copied_dir = DirectoryItem(self.Filesystem.dir_copyer.copy_directory(dir_to_copy))
        try:
            assert new_copied_dir not in items_before, f"copied `{new_copied_dir}` already exists"

            items_after = self.Filesystem.navigator.get_all_list_items()

            assert len(items_before.dirs) + 1 == len(items_after.dirs), \
                (f"expected {len(items_before.dirs) + 1} directories, but got {len(items_after.dirs)}: "
                 f"{items_after.dirs}")

            assert items_before.files == items_after.files, \
                f"different files detected - before {items_before.files} / after {items_after.files}"

            remaining_dirs = items_after.dirs.subtract(items_before.dirs)

            assert len(remaining_dirs) == 1, \
                f"more than one directories have changed: {remaining_dirs}"

            assert remaining_dirs[0] == new_copied_dir, \
                f"the name of the changed directory does not match: `{remaining_dirs[0]}` (expected `{new_copied_dir}`)"
        finally:
            self.Filesystem.dir_deleter.delete_directory(new_copied_dir)



import pathlib
from typing import Union

import balder

from balderhub.fileops.lib.utils import DirectoryItem


class DirectoryMoverFeature(balder.Feature):
    """
    Feature that allows to move a directory within the filesystem
    """
    def move_directory(self, source: DirectoryItem, destination: Union[DirectoryItem, pathlib.Path]) -> None:
        """
        This method moves a directory (provided by ``source``) to the destination directory.

        :param source: the directory to be moved
        :param destination: the destination directory
        """
        raise NotImplementedError()



import balder

from balderhub.fileops.scenarios.abstract_scenario_base import AbstractScenarioBase
from balderhub.fileops.lib.scenario_features import FileSystemNavigatorFeature, DirectoryCreatorFeature, \
    DirectoryDeleterFeature
from balderhub.fileops.lib.utils import DirectoryItem


class ScenarioDirCreateStandalone(AbstractScenarioBase):
    """
    Standalone scenario that provides test for CREATING directories.

    This standalone version also cleans up everything afterward.
    """

    class Filesystem(AbstractScenarioBase.Filesystem):
        """
        The main device that has the filesystem by itself or allows to control it
        """
        navigator = FileSystemNavigatorFeature()
        dir_creator = DirectoryCreatorFeature()
        dir_deleter = DirectoryDeleterFeature()  # TODO another scenario with already provided dirs??


    @balder.fixture('testcase')
    def dir_to_create(self):
        """fixture that creates and cleans up a test directory"""
        directory = DirectoryItem('dir_to_create')
        return directory

    def test_create_new_valid_dir(self, dir_to_create):
        """
        Simple test that tries to create a test directory. It will automatically clean up everything it changed
        afterward.
        """
        self.Filesystem.navigator.navigate_to(self.EXECUTE_IN_DIR)

        items_before = self.Filesystem.navigator.get_all_list_items()

        assert dir_to_create not in items_before, f"the {dir_to_create} already exists - can not create it again"

        try:
            self.Filesystem.dir_creator.create_new_directory(dir_to_create.name)
            items_after = self.Filesystem.navigator.get_all_list_items()

            assert items_before.files.compare_with(items_after.files), \
                f"different files detected - before {items_before.files} / after {items_after.files}"

            assert len(items_after.dirs) == len(items_before.dirs) + 1, \
                (f"detect unexpected number of directories: {len(items_after.dirs)} "
                 f"(expected {len(items_before.dirs) + 1}) - "
                 f"before: {items_before.dirs} | after: {items_after.dirs}")

            new_items = items_after.subtract(items_before)

            assert len(new_items) == 1, \
                f"detect unexpected number of added directories: {new_items} (expected only `{dir_to_create}`)"
            assert new_items[0] == dir_to_create, \
                f"detect unexpected directory: {new_items[0]} (expected {dir_to_create})"
            # TODO also check the content

        finally:
            self.Filesystem.dir_deleter.delete_directory(dir_to_create)



from .directory_item import DirectoryItem
from .file_item import FileItem
from .file_system_item import FileSystemItem
from .file_system_item_list import FileSystemItemList

__all__ = [
    "DirectoryItem",
    "FileItem",
    "FileSystemItem",
    "FileSystemItemList",
]



import balder

from balderhub.fileops.scenarios.abstract_scenario_base import AbstractScenarioBase
from balderhub.fileops.lib.scenario_features import FileSystemNavigatorFeature, DirectoryCreatorFeature, \
    DirectoryDeleterFeature, DirectoryRenamerFeature
from balderhub.fileops.lib.utils import DirectoryItem


class ScenarioDirRenameStandalone(AbstractScenarioBase):
    """
    Standalone scenario that provides test for RENAMING directories.

    This standalone version also creates a test directory to rename and cleans up everything afterward.
    """

    class Filesystem(AbstractScenarioBase.Filesystem):
        """
        The main device that has the filesystem by itself or allows to control it
        """
        navigator = FileSystemNavigatorFeature()
        dir_creator = DirectoryCreatorFeature()    # TODO another scenario with already provided dirs??
        dir_deleter = DirectoryDeleterFeature()    # TODO another scenario with already provided dirs??
        dir_rename = DirectoryRenamerFeature()

    @balder.fixture('testcase')
    def dir_to_rename(self):
        """fixture that creates and cleans up a test directory"""
        directory = DirectoryItem('dir_to_rename')
        self.Filesystem.dir_creator.create_new_directory(directory.name)
        yield directory
        if directory in self.Filesystem.navigator.get_all_list_items():
            self.Filesystem.dir_deleter.delete_directory(directory)

    def test_rename_directory(self, dir_to_rename):
        """
        Simple test that tries to rename the test directory. It will automatically clean up everything it changed
        afterward.
        """
        items_before = self.Filesystem.navigator.get_all_list_items()

        # now copy the first directory
        new_renamed_dir = DirectoryItem(f"{dir_to_rename.name}_renamed")

        assert new_renamed_dir not in items_before, \
            f"a`{new_renamed_dir}` already exists within elements {items_before}"

        try:
            self.Filesystem.dir_rename.rename_directory(dir_to_rename, new_renamed_dir.name)

            items_after = self.Filesystem.navigator.get_all_list_items()

            assert items_before.files == items_after.files, \
                f"different files detected - before {items_before.files} / after {items_after.files}"

            assert len(items_before.dirs) == len(items_after.dirs), \
                f"directory count should not change - before: {items_before.dirs} / after: {items_after.dirs}"

            assert dir_to_rename not in items_after.dirs, \
                f"{dir_to_rename} should not be available in file list after renaming: {items_after.dirs}"

            assert new_renamed_dir in items_after.dirs, \
                f"{new_renamed_dir} should be available  in file list after renaming: {items_after.dirs}"

            # todo make sure that content of `dir_to_move` is in `new_moved_dir`
        finally:
            if new_renamed_dir in self.Filesystem.navigator.get_all_list_items():
                self.Filesystem.dir_deleter.delete_directory(new_renamed_dir)



from .scenario_dir_copy import ScenarioDirCopyStandalone
from .scenario_dir_create import ScenarioDirCreateStandalone
from .scenario_dir_delete import ScenarioDirDeleteStandalone
from .scenario_dir_move import ScenarioDirMoveStandalone
from .scenario_dir_rename import ScenarioDirRenameStandalone


__all__ = [
    'ScenarioDirCopyStandalone',
    'ScenarioDirCreateStandalone',
    'ScenarioDirDeleteStandalone',
    'ScenarioDirMoveStandalone',
    'ScenarioDirRenameStandalone',
]



import balder

from balderhub.fileops.lib.utils import DirectoryItem


class DirectoryCopyerFeature(balder.Feature):
    """
    Feature that allows to copy a directory within the filesystem
    """

    def copy_directory(self, source: DirectoryItem) -> str:
        """
        This method copies a directory and returns the new name of the directory.

        :param source: the directory that should be copied
        :return: the name of the copied directory
        """
        raise NotImplementedError()



import balder

from balderhub.fileops.lib.utils import DirectoryItem


class DirectoryDeleterFeature(balder.Feature):
    """
    Feature that allows to delete a directory within the filesystem
    """

    def delete_directory(self, directory: DirectoryItem) -> None:
        """
        This method deletes a directory from the filesystem.

        :param directory: the directory to delete
        """
        raise NotImplementedError()



import pathlib

import balder

from ..utils.file_system_item_list import FileSystemItemList


class FileSystemNavigatorFeature(balder.Feature):
    """
    Feature that allows to navigate in the filesystem and allows to returns a list of the elements of the current
    active path
    """
    def navigate_to(self, path: pathlib.Path) -> None:
        """
        This method navigates to the specified path

        :param path: the path in the filesystem to navigate to
        """
        raise NotImplementedError()

    def get_all_list_items(self) -> FileSystemItemList:
        """
        This method returns a list of the files/directories at the current path in the filesystem

        :return: a :class:`FileSystemItemList` list with all items at the current path in the filesystem
        """
        raise NotImplementedError()



import pathlib
import balder

from balderhub.fileops.lib.scenario_features import FileSystemNavigatorFeature


class AbstractScenarioBase(balder.Scenario):
    """
    Abstract scenario class that is shared for almost all scenarios within this project
    """
    #: this is the directory within this scenario should execute its tests
    EXECUTE_IN_DIR = pathlib.Path('/')

    class Filesystem(balder.Device):
        """
        The main device that has the filesystem by itself or allows to control it
        """
        navigator = FileSystemNavigatorFeature()
